@@ -36,7 +36,9 @@ import base64
 import hashlib
 import logging
 import re
+import sys
 import time
+import binascii
 
 # only needed for arc
 try:
@@ -70,8 +72,13 @@ from dkim.crypto import (
 try:
     from dkim.dnsplug import get_txt
 except ImportError:
-    def get_txt(s):
-        raise RuntimeError("DKIM.verify requires DNS or dnspython module")
+    try:
+        import aiodns
+        from dkim.asyncsupport import get_txt_async as get_txt
+    except:
+        # Only true if not using async
+        def get_txt(s,timeout=5):
+            raise RuntimeError("DKIM.verify requires DNS or dnspython module")
 from dkim.util import (
     get_default_logger,
     InvalidTagValueList,
@@ -419,8 +426,7 @@ def fold(header, namelen=0, linesep=b'\r\n'):
             return pre + header
 
 
-def load_pk_from_dns(name, dnsfunc=get_txt):
-  s = dnsfunc(name)
+def evaluate_pk(name, s):
   if not s:
       raise KeyFormatError("missing public key: %s"%name)
   try:
@@ -453,17 +459,26 @@ def load_pk_from_dns(name, dnsfunc=get_txt):
       ktag = b'rsa'
   if pub[b'k'] != b'rsa' and pub[b'k'] != b'ed25519':
       raise KeyFormatError('unknown algorithm in k= tag: {0}'.format(pub[b'k']))
+  seqtlsrpt = False
   try:
       # Ignore unknown service types, RFC 6376 3.6.1
-      if pub[b's'] != b'*' and pub[b's'] != b'email':
+      if pub[b's'] != b'*' and pub[b's'] != b'email' and pub[b's'] != b'tlsrpt':
           pk = None
           keysize = None
           ktag = None
           raise KeyFormatError('unknown service type in s= tag: {0}'.format(pub[b's']))
+      elif pub[b's'] == b'tlsrpt':
+          seqtlsrpt = True
   except:
       # Default is '*' - all service types, so no error if missing from key record
       pass
-  return pk, keysize, ktag
+  return pk, keysize, ktag, seqtlsrpt
+
+
+def load_pk_from_dns(name, dnsfunc=get_txt, timeout=5):
+  s = dnsfunc(name, timeout=timeout)
+  pk, keysize, ktag, seqtlsrpt = evaluate_pk(name, s)
+  return pk, keysize, ktag, seqtlsrpt
 
 
 #: Abstract base class for holding messages and options during DKIM/ARC signing and verification.
@@ -477,8 +492,13 @@ class DomainSigner(object):
   #: @param signature_algorithm: the signing algorithm to use when signing
   #: @param debug_content: log headers and body after canonicalization (default False)
   #: @param linesep: use this line seperator for folding the headers
+  #: @param timeout: number of seconds for DNS lookup timeout (default = 5)
+  #: @param tlsrpt: message is an RFC 8460 TLS report (default False)
+  #: False: Not a tlsrpt, True: Is a tlsrpt, 'strict': tlsrpt, invalid if
+  #: service type is missing. For signing, if True, length is never used.
   def __init__(self,message=None,logger=None,signature_algorithm=b'rsa-sha256',
-        minkey=1024, linesep=b'\r\n', debug_content=False):
+        minkey=1024, linesep=b'\r\n', debug_content=False, timeout=5,
+        tlsrpt=False):
     self.set_message(message)
     if logger is None:
         logger = get_default_logger()
@@ -502,6 +522,10 @@ class DomainSigner(object):
     self.minkey = minkey
     # use this line seperator for output
     self.linesep = linesep
+    self.timeout = timeout
+    self.tlsrpt = tlsrpt
+    # Service type in DKIM record is s=tlsrpt
+    self.seqtlsrpt = False
 
 
   #: Header fields to protect from additions by default.
@@ -663,18 +687,15 @@ class DomainSigner(object):
 
     return header_value
 
-  # Abstract helper method to verify a signed header
-  #: @param sig: List of (key, value) tuples containing tag=values of the header
-  #: @param include_headers: headers to validate b= signature against
-  #: @param sig_header: (header_name, header_value)
-  #: @param dnsfunc: interface to dns
-  def verify_sig(self, sig, include_headers, sig_header, dnsfunc):
-    name = sig[b's'] + b"._domainkey." + sig[b'd'] + b"."
-    try:
-      pk, self.keysize, ktag = load_pk_from_dns(name, dnsfunc)
-    except KeyFormatError as e:
-      self.logger.error("%s" % e)
-      return False
+  def verify_sig_process(self, sig, include_headers, sig_header, dnsfunc):
+    """Non-async sensitive verify_sig elements.  Separated to avoid async code
+    duplication."""
+    # RFC 8460 MAY ignore signatures without tlsrpt Service Type
+    if self.tlsrpt == 'strict' and not self.seqtlsrpt:
+        raise ValidationError("Message is tlsrpt and Service Type is not tlsrpt")
+    # Inferred requirement from both RFC 8460 and RFC 6376
+    if not self.tlsrpt and self.seqtlsrpt:
+        raise ValidationError("Message is not tlsrpt and Service Type is tlsrpt")
 
     try:
         canon_policy = CanonicalizationPolicy.from_c_value(sig.get(b'c', b'simple/simple'))
@@ -688,7 +709,7 @@ class DomainSigner(object):
       h = HashThrough(hasher(), self.debug_content)
 
       body = canon_policy.canonicalize_body(self.body)
-      if b'l' in sig:
+      if b'l' in sig and not self.tlsrpt:
         body = body[:int(sig[b'l'])]
       h.update(body)
       if self.debug_content:
@@ -719,24 +740,43 @@ class DomainSigner(object):
     if self.debug_content:
         self.logger.debug("signed for %s: %r" % (sig_header[0], h.hashed()))
     signature = base64.b64decode(re.sub(br"\s+", b"", sig[b'b']))
-    if ktag == b'rsa':
+    if self.ktag == b'rsa':
         try:
-            res = RSASSA_PKCS1_v1_5_verify(h, signature, pk)
+            res = RSASSA_PKCS1_v1_5_verify(h, signature, self.pk)
             self.logger.debug("%s valid: %s" % (sig_header[0], res))
             if res and self.keysize < self.minkey:
                 raise KeyFormatError("public key too small: %d" % self.keysize)
             return res
         except (TypeError,DigestTooLargeError) as e:
             raise KeyFormatError("digest too large for modulus: %s"%e)
-    elif ktag == b'ed25519':
+    elif self.ktag == b'ed25519':
         try:
-            pk.verify(h.digest(), signature)
+            self.pk.verify(h.digest(), signature)
             self.logger.debug("%s valid" % (sig_header[0]))
             return True
         except (nacl.exceptions.BadSignatureError) as e:
             return False
     else:
-        raise UnknownKeyTypeError(ktag)
+        raise UnknownKeyTypeError(self.ktag)
+
+
+  # Abstract helper method to verify a signed header
+  #: @param sig: List of (key, value) tuples containing tag=values of the header
+  #: @param include_headers: headers to validate b= signature against
+  #: @param sig_header: (header_name, header_value)
+  #: @param dnsfunc: interface to dns
+  def verify_sig(self, sig, include_headers, sig_header, dnsfunc):
+    name = sig[b's'] + b"._domainkey." + sig[b'd'] + b"."
+    try:
+      self.pk, self.keysize, self.ktag, self.seqtlsrpt = load_pk_from_dns(name,
+              dnsfunc, timeout=self.timeout)
+    except KeyFormatError as e:
+      self.logger.error("%s" % e)
+      return False
+    except binascii.Error as e:
+      self.logger.error('KeyFormatError: {0}'.format(e))
+      return False
+    return self.verify_sig_process(sig, include_headers, sig_header, dnsfunc)
 
 
 #: Hold messages and options during DKIM signing and verification.
@@ -806,6 +846,10 @@ class DKIM(DomainSigner):
     # record what verify should extract
     self.include_headers = include_headers
 
+    if self.tlsrpt:
+        # RFC 8460 MUST NOT
+        length = False
+
     # rfc4871 says FROM is required
     if b'from' not in include_headers:
         raise ParameterError("The From header field MUST be signed")
@@ -847,15 +891,14 @@ class DKIM(DomainSigner):
     self.signature_fields = dict(sigfields)
     return b'DKIM-Signature: ' + res
 
-  #: Verify a DKIM signature.
-  #: @type idx: int
-  #: @param idx: which signature to verify.  The first (topmost) signature is 0.
-  #: @type dnsfunc: callable
-  #: @param dnsfunc: an option function to lookup TXT resource records
-  #: for a DNS domain.  The default uses dnspython or pydns.
-  #: @return: True if signature verifies or False otherwise
-  #: @raise DKIMException: when the message, signature, or key are badly formed
-  def verify(self,idx=0,dnsfunc=get_txt):
+  #: Checks if any DKIM signature is present
+  #: @return: True if there is one or more DKIM signatures present or False otherwise
+  def present(self):
+    return (len([(x,y) for x,y in self.headers if x.lower() == b"dkim-signature"]) > 0)
+
+  def verify_headerprep(self, idx=0):
+    """Non-DNS verify parts to minimize asyncio code duplication."""
+
     sigheaders = [(x,y) for x,y in self.headers if x.lower() == b"dkim-signature"]
     if len(sigheaders) <= idx:
         return False
@@ -875,7 +918,18 @@ class DKIM(DomainSigner):
 
     include_headers = [x.lower() for x in re.split(br"\s*:\s*", sig[b'h'])]
     self.include_headers = tuple(include_headers)
+    return sig, include_headers, sigheaders
 
+  #: Verify a DKIM signature.
+  #: @type idx: int
+  #: @param idx: which signature to verify.  The first (topmost) signature is 0.
+  #: @type dnsfunc: callable
+  #: @param dnsfunc: an option function to lookup TXT resource records
+  #: for a DNS domain.  The default uses dnspython or pydns.
+  #: @return: True if signature verifies or False otherwise
+  #: @raise DKIMException: when the message, signature, or key are badly formed
+  def verify(self,idx=0,dnsfunc=get_txt):
+    sig, include_headers, sigheaders = self.verify_headerprep(idx=0)
     return self.verify_sig(sig, include_headers, sigheaders[idx], dnsfunc)
 
 
@@ -1248,7 +1302,7 @@ def sign(message, selector, domain, privkey, identity=None,
          canonicalize=(b'relaxed', b'simple'),
          signature_algorithm=b'rsa-sha256',
          include_headers=None, length=False, logger=None,
-         linesep=b'\r\n'):
+         linesep=b'\r\n', tlsrpt=False):
     # type: (bytes, bytes, bytes, bytes, bytes, tuple, bytes, list, bool, any) -> bytes
     """Sign an RFC822 message and return the DKIM-Signature header line.
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
@@ -1262,22 +1316,30 @@ def sign(message, selector, domain, privkey, identity=None,
     @param length: true if the l= tag should be included to indicate body length (default False)
     @param logger: a logger to which debug info will be written (default None)
     @param linesep: use this line seperator for folding the headers
+    @param tlsrpt: message is an RFC 8460 TLS report (default False)
+     False: Not a tlsrpt, True: Is a tlsrpt, 'strict': tlsrpt, invalid if
+     service type is missing. For signing, if True, length is never used.
     @return: DKIM-Signature header field terminated by \\r\\n
     @raise DKIMException: when the message, include_headers, or key are badly formed.
     """
 
-    d = DKIM(message,logger=logger,signature_algorithm=signature_algorithm,linesep=linesep)
+    d = DKIM(message,logger=logger,signature_algorithm=signature_algorithm,linesep=linesep,tlsrpt=tlsrpt)
     return d.sign(selector, domain, privkey, identity=identity, canonicalize=canonicalize, include_headers=include_headers, length=length)
 
 
-def verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
+def verify(message, logger=None, dnsfunc=get_txt, minkey=1024,
+        timeout=5, tlsrpt=False):
     """Verify the first (topmost) DKIM signature on an RFC822 formatted message.
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param logger: a logger to which debug info will be written (default None)
+    @param timeout: number of seconds for DNS lookup timeout (default = 5)
+    @param tlsrpt: message is an RFC 8460 TLS report (default False)
+     False: Not a tlsrpt, True: Is a tlsrpt, 'strict': tlsrpt, invalid if
+     service type is missing. For signing, if True, length is never used.
     @return: True if signature verifies or False otherwise
     """
     # type: (bytes, any, function, int) -> bool
-    d = DKIM(message,logger=logger,minkey=minkey)
+    d = DKIM(message,logger=logger,minkey=minkey,timeout=timeout,tlsrpt=tlsrpt)
     try:
         return d.verify(dnsfunc=dnsfunc)
     except DKIMException as x:
@@ -1285,6 +1347,18 @@ def verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
             logger.error("%s" % x)
         return False
 
+
+# aiodns requires Python 3.5+, so no async before that
+if sys.version_info >= (3, 5):
+    try:
+        import aiodns
+        from dkim.asyncsupport import verify_async
+        dkim_verify_async = verify_async
+    except ImportError:
+        # If aiodns is not installed, then async verification is not available
+        pass
+
+
 # For consistency with ARC
 dkim_sign = sign
 dkim_verify = verify
@@ -1317,17 +1391,18 @@ def arc_sign(message, selector, domain, privkey,
                   timestamp=timestamp, standardize=standardize)
 
 
-def arc_verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
+def arc_verify(message, logger=None, dnsfunc=get_txt, minkey=1024, timeout=5):
     # type: (bytes, any, function, int) -> tuple
     """Verify the ARC chain on an RFC822 formatted message.
     @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param logger: a logger to which debug info will be written (default None)
     @param dnsfunc: an optional function to lookup TXT resource records
     @param minkey: the minimum key size to accept
+    @param timeout: number of seconds for DNS lookup timeout (default = 5)
     @return: three-tuple of (CV Result (CV_Pass, CV_Fail or CV_None), list of
     result dictionaries, result reason)
     """
-    a = ARC(message,logger=logger,minkey=minkey)
+    a = ARC(message,logger=logger,minkey=minkey,timeout=5)
     try:
         return a.verify(dnsfunc=dnsfunc)
     except DKIMException as x:
